@@ -18,20 +18,22 @@ along with this program if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 """
 
-from PyQt4.QtCore import Qt, QPointF, QPoint, QRectF, QMimeData, QSize
+from PyQt4.QtCore import Qt, QPointF, QPoint, QRectF, QMimeData, QSize, QVariant
 from PyQt4.QtGui import QGraphicsRectItem, QGraphicsItem, QSizePolicy, QFrame, QFont
 from PyQt4.QtGui import QGraphicsView, QGraphicsEllipseItem, QGraphicsScene, QLabel
 from PyQt4.QtGui import QColor, QPainter, QDrag, QPixmap, QStyleOptionGraphicsItem, QPen, QBrush
 from PyQt4.QtGui import QFontMetrics, QTransform
+from PyQt4.QtGui import QMenu, QCursor
 from PyQt4.QtSvg import QGraphicsSvgItem
 from tileset import Tileset, TileException
-from tile import Tile, chiNext, GraphicsTileItem
+from tile import chiNext, Tile, elements
+from uitile import UITile, UIMeld
 from meld import Meld
 from animation import Animation, Animated, animate
 from message import Message
 
-from util import logDebug, logException, m18nc, isAlive, kprint, stack, uniqueList
-from common import elements, WINDS, LIGHTSOURCES, InternalParameters, ZValues, Debug, Preferences
+from util import logDebug, logException, m18n, m18nc, kprint, stack, uniqueList
+from common import WINDS, LIGHTSOURCES, Internal, ZValues, Debug, Preferences, isAlive
 
 ROUNDWINDCOLOR = QColor(235, 235, 173)
 
@@ -113,18 +115,17 @@ class PlayerWind(QGraphicsEllipseItem):
 class WindLabel(QLabel):
     """QLabel holding the wind tile"""
 
-    @apply
-    def wind(): # pylint: disable=E0202
+    @property
+    def wind(self):
+        """the current wind on this label"""
+        return self.__wind
+
+    @wind.setter
+    def wind(self, wind):
         """setting the wind also changes the pixmap"""
-        def fget(self):
-            # pylint: disable=W0212
-            return self.__wind
-        def fset(self, wind):
-            # pylint: disable=W0212
-            if self.__wind != wind:
-                self.__wind = wind
-                self._refresh()
-        return property(**locals())
+        if self.__wind != wind:
+            self.__wind = wind
+            self._refresh()
 
     def __init__(self, wind = None, roundsFinished = 0, parent=None):
         QLabel.__init__(self, parent)
@@ -134,18 +135,17 @@ class WindLabel(QLabel):
         self.__roundsFinished = roundsFinished
         self.wind = wind
 
-    @apply
-    def roundsFinished():
+    @property
+    def roundsFinished(self):
         """setting roundsFinished also changes pixmaps if needed"""
-        def fget(self):
-            # pylint: disable=W0212
-            return self.__roundsFinished
-        def fset(self, roundsFinished):
-            # pylint: disable=W0212
-            if self.__roundsFinished != roundsFinished:
-                self.__roundsFinished = roundsFinished
-                self._refresh()
-        return property(**locals())
+        return self.__roundsFinished
+
+    @roundsFinished.setter
+    def roundsFinished(self, roundsFinished):
+        """setting roundsFinished also changes pixmaps if needed"""
+        if self.__roundsFinished != roundsFinished:
+            self.__roundsFinished = roundsFinished
+            self._refresh()
 
     def _refresh(self):
         """update pixmaps"""
@@ -155,8 +155,7 @@ class WindLabel(QLabel):
 
 class Board(QGraphicsRectItem):
     """ a board with any number of positioned tiles"""
-    # pylint: disable=R0902
-    # pylint we need more than 10 instance attributes
+    # pylint: disable=too-many-instance-attributes
 
     arrows = [Qt.Key_Left, Qt.Key_Down, Qt.Key_Up, Qt.Key_Right]
     def __init__(self, width, height, tileset, boardRotation=0):
@@ -164,6 +163,7 @@ class Board(QGraphicsRectItem):
         self.tiles = []
         self.isHandBoard = False
         self._focusTile = None
+        self.__prevPos = 0
         self._noPen()
         self.tileDragEnabled = False
         self.setRotation(boardRotation)
@@ -179,46 +179,64 @@ class Board(QGraphicsRectItem):
         self.tileset = tileset
         self.level = 0
 
-    # pylint: disable=R0201
+    # pylint: disable=no-self-use
     def name(self):
         """default board name, used for debugging messages"""
         return 'board'
 
+    def hide(self):
+        """remove all tile references so they can be garbage collected"""
+        for tile in self.tiles:
+            tile.hide()
+        self.tiles = []
+        self._focusTile = None
+        if isAlive(self):
+            QGraphicsRectItem.hide(self)
+
     def autoSelectTile(self):
         """call this when kajongg should automatically focus
         on an appropriate tile"""
-        focusableTiles = self._focusableTiles()
-        if len(focusableTiles):
-            return focusableTiles[0]
+        focusCandidates = self._focusableTiles()
+        if focusCandidates:
+            firstCandidate = focusCandidates[0]
+            if self._focusTile not in focusCandidates:
+                focusCandidates = list(x for x in focusCandidates if x.sortKey() >= self.__prevPos)
+                focusCandidates.append(firstCandidate)
+                self.focusTile = focusCandidates[0]
 
-    @apply
-    def focusTile(): # pylint: disable=E0202
+    @property
+    def currentFocusTile(self):
+        """get focusTile without selecting one"""
+        return self._focusTile
+
+    @property
+    def focusTile(self):
         """the tile of this board with focus. This is per Board!"""
-        def fget(self):
-            # pylint: disable=W0212
-            if self._focusTile is None:
-                self._focusTile = self.autoSelectTile()
-            return self._focusTile
-        def fset(self, tile):
-            # pylint: disable=W0212
-            prevTile = self._focusTile
-            if tile:
-                assert tile.element != 'Xy', tile
-                if not isinstance(tile.board, DiscardBoard):
-                    assert tile.focusable, tile
-                self._focusTile = tile
-            else:
-                self._focusTile = self.autoSelectTile()
-            if self._focusTile and self._focusTile.element in Debug.focusable:
-                logDebug('new focus tile %s from %s' % (self._focusTile.element, stack('')[-1]))
-            if (self._focusTile != prevTile
-                and self.isHandBoard and self.player
-                and not self.player.game.isScoringGame()
-                and InternalParameters.field.clientDialog):
-                InternalParameters.field.clientDialog.focusTileChanged()
-            if self.hasFocus:
-                self.scene().focusBoard = self
-        return property(**locals())
+        if self._focusTile is None:
+            self.autoSelectTile()
+        return self._focusTile
+
+    @focusTile.setter
+    def focusTile(self, tile):
+        """the tile of this board with focus. This is per Board!"""
+        if tile is self._focusTile:
+            return
+        if tile:
+            assert tile.tile != 'Xy', tile
+            if not isinstance(tile.board, DiscardBoard):
+                assert tile.focusable, tile
+            self.__prevPos = tile.sortKey()
+        self._focusTile = tile
+        if self._focusTile and self._focusTile.tile in Debug.focusable:
+            logDebug('%s: new focus tile %s from %s' % (
+                self.name(), self._focusTile.tile if self._focusTile else 'None', stack('')[-1]))
+        if (self.isHandBoard and self.player
+            and not self.player.game.isScoringGame()
+            and Internal.field.clientDialog):
+            Internal.field.clientDialog.focusTileChanged()
+        if self.hasFocus:
+            self.scene().focusBoard = self
+        assert tile is self._focusTile
 
     def setEnabled(self, enabled):
         """enable/disable this board"""
@@ -231,32 +249,24 @@ class Board(QGraphicsRectItem):
         the next list element.
         respect board orientation: Right Arrow should always move right
         relative to the screen, not relative to the board"""
-        dirs = [Qt.Key_Right, Qt.Key_Down, Qt.Key_Left, Qt.Key_Up] * 2
-        sorter = dirs[dirs.index(sortDir) + self.sceneRotation()//90]
-        if sorter == Qt.Key_Down:
-            sortFunction = lambda x: x.xoffset * 100 + x.yoffset
-        elif sorter == Qt.Key_Up:
-            sortFunction = lambda x: -x.xoffset * 100 - x.yoffset
-        elif sorter == Qt.Key_Left:
-            sortFunction = lambda x: -x.yoffset * 100 - x.xoffset
-        else:
-            sortFunction = lambda x: x.yoffset * 100 + x.xoffset
-        return sorted([x for x in self.tiles if x.focusable], key=sortFunction)
+        return sorted([x for x in self.tiles if x.focusable], key=lambda x: x.sortKey(sortDir))
 
-    @apply
-    def hasFocus():
+    @property
+    def hasFocus(self):
         """defines if this board should show a focusRect
         if another board has focus, setting this to False does
         not change scene.focusBoard"""
-        def fget(self):
-            # pylint: disable=W0212
-            return self.scene() and self.scene().focusBoard == self
-        def fset(self, value):
-            # pylint: disable=W0212
+        return self.scene() and self.scene().focusBoard == self and self._focusTile
+
+    @hasFocus.setter
+    def hasFocus(self, value):
+        """sets focus on this board"""
+        if isAlive(self):
             scene = self.scene()
             if scene.focusBoard == self or value:
+                if self.focusTile:
+                    assert self.focusTile.board == self, '%s not in self %s' % (self.focusTile, self)
                 scene.focusBoard = self if value else None
-        return property(**locals())
 
     @staticmethod
     def mapChar2Arrow(event):
@@ -291,11 +301,36 @@ class Board(QGraphicsRectItem):
             tiles.append(tiles[0])
             self.focusTile = tiles[tiles.index(self.focusTile)+1]
 
-    def dragObject(self, tile):
-        """returns the object that should be dragged when the user tries to drag
-        tile. This is either only the tile or the meld containing this tile"""
-        # pylint: disable=R0201
-        return tile, None
+    def uiMeldWithTile(self, uiTile):
+        """returns the UI Meld with uiTile. A Board does not know about melds,
+        so default is to return a Meld with only uiTile"""
+        return Meld(uiTile)
+
+    def meldVariants(self, uiTile, lowerHalf): # pylint: disable=unused-argument
+        """all possible melds that could be meant by dragging/dropping tile"""
+        return [Meld(uiTile)]
+
+    def chooseVariant(self, uiTile, lowerHalf=False):
+        """make the user choose from a list of possible melds for the target.
+        The melds do not contain real Tiles, just the scoring strings."""
+        variants = self.meldVariants(uiTile, lowerHalf)
+        idx = 0
+        if len(variants) > 1:
+            menu = QMenu(m18n('Choose from'))
+            for idx, variant in enumerate(variants):
+                action = menu.addAction(variant.typeName())
+                action.setData(QVariant(idx))
+            if Internal.field.centralView.dragObject:
+                menuPoint = QCursor.pos()
+            else:
+                menuPoint = uiTile.board.tileFaceRect().bottomRight()
+                view = Internal.field.centralView
+                menuPoint = view.mapToGlobal(view.mapFromScene(uiTile.mapToScene(menuPoint)))
+            action = menu.exec_(menuPoint)
+            if not action:
+                return None
+            idx = action.data().toInt()[0]
+        return variants[idx]
 
     def dragEnterEvent(self, dummyEvent):
         """drag enters the HandBoard: highlight it"""
@@ -317,7 +352,7 @@ class Board(QGraphicsRectItem):
 
     def tilesByElement(self, element):
         """returns all child items holding a tile for element"""
-        return list(x for x in self.tiles if x.element == element)
+        return list(x for x in self.tiles if x.tile == element)
 
     def rotatedLightSource(self):
         """the light source we need for the original tile before it is rotated"""
@@ -401,45 +436,41 @@ class Board(QGraphicsRectItem):
         newY = self.yWidth*width+self.yHeight*height + offsets[1]
         QGraphicsRectItem.setPos(self, newX, newY)
 
-    @apply
-    def showShadows():
+    @property
+    def showShadows(self):
         """the active lightSource"""
-        def fget(self):
-            # pylint: disable=W0212
-            return self._showShadows
-        def fset(self, value):
-            """set active lightSource"""
-            # pylint: disable=W0212
-            if self._showShadows != value:
-                for tile in self.tiles:
-                    tile.graphics.setClippingFlags()
-                self._reload(self.tileset, showShadows=value)
-        return property(**locals())
+        return self._showShadows
 
-    @apply
-    def lightSource():
+    @showShadows.setter
+    def showShadows(self, value):
+        """set active lightSource"""
+        if self._showShadows != value:
+            for tile in self.tiles:
+                tile.setClippingFlags()
+            self._reload(self.tileset, showShadows=value)
+
+    @property
+    def lightSource(self):
         """the active lightSource"""
-        def fget(self):
-            # pylint: disable=W0212
-            return self._lightSource
-        def fset(self, lightSource):
-            """set active lightSource"""
-            # pylint: disable=W0212
-            if self._lightSource != lightSource:
-                if lightSource not in LIGHTSOURCES:
-                    logException(TileException('lightSource %s illegal' % lightSource))
-                self._reload(self.tileset, lightSource)
-        return property(**locals())
+        return self._lightSource
 
-    @apply
-    def tileset(): # pylint: disable=E0202
+    @lightSource.setter
+    def lightSource(self, lightSource):
+        """set active lightSource"""
+        if self._lightSource != lightSource:
+            if lightSource not in LIGHTSOURCES:
+                logException(TileException('lightSource %s illegal' % lightSource))
+            self._reload(self.tileset, lightSource)
+
+    @property
+    def tileset(self):
         """get/set the active tileset and resize accordingly"""
-        def fget(self):
-            # pylint: disable=W0212
-            return self._tileset
-        def fset(self, tileset):
-            self._reload(tileset, self._lightSource) # pylint: disable=W0212
-        return property(**locals())
+        return self._tileset
+
+    @tileset.setter
+    def tileset(self, tileset):
+        """get/set the active tileset and resize accordingly"""
+        self._reload(tileset, self._lightSource)
 
     def _reload(self, tileset=None, lightSource=None, showShadows=None):
         """call this if tileset or lightsource change: recomputes the entire board"""
@@ -462,12 +493,12 @@ class Board(QGraphicsRectItem):
                     child.showShadows = showShadows
             for tile in self.tiles:
                 self.placeTile(tile)
-                tile.graphics.update()
+                tile.update()
             self.computeRect()
             if self.hasFocus:
                 self.scene().focusBoard = self
 
-    def focusRectWidth(self): # pylint: disable=R0201
+    def focusRectWidth(self): # pylint: disable=no-self-use
         """how many tiles are in focus rect?"""
         return 1
 
@@ -505,20 +536,20 @@ class Board(QGraphicsRectItem):
     def __tilePlace(self, tile):
         """compute all properties for tile in this board: pos, scale, rotation
         and return them in a dict"""
-        assert isinstance(tile, Tile)
-        if not tile.graphics.scene():
-            self.scene().addItem(tile.graphics)
+        assert isinstance(tile, UITile)
+        if not tile.scene():
+            self.scene().addItem(tile)
         width = self.tileset.faceSize.width()
         height = self.tileset.faceSize.height()
         shiftZ = self.shiftZ(tile.level)
         boardPos = QPointF(tile.xoffset*width, tile.yoffset*height) + shiftZ
         scenePos = self.mapToScene(boardPos)
-        tile.graphics.setDrawingOrder()
+        tile.setDrawingOrder()
         return {'pos': scenePos, 'rotation': self.sceneRotation(), 'scale': self.scale()}
 
     def placeTile(self, tile):
         """places the tile in the scene. With direct=False, animate"""
-        assert isinstance(tile, Tile)
+        assert isinstance(tile, UITile)
         for pName, newValue in self.__tilePlace(tile).items():
             animation = tile.queuedAnimation(pName)
             if animation:
@@ -536,101 +567,121 @@ class Board(QGraphicsRectItem):
                     if curValue != newValue:
                         Animation(tile, pName, newValue)
 
+    def addUITile(self, uiTile):
+        """add uiTile to this board"""
+        self.tiles.append(uiTile)
+
+    def removeUITile(self, uiTile):
+        """remove uiTile from this board"""
+        self.tiles.remove(uiTile)
+        if self.currentFocusTile == uiTile:
+            self.focusTile = None
+
 class CourtBoard(Board):
     """A Board that is displayed within the wall"""
     def __init__(self, width, height):
-        Board.__init__(self, width, height, InternalParameters.field.tileset)
+        Board.__init__(self, width, height, Internal.field.tileset)
+        self.setAcceptDrops(True)
 
     def maximize(self):
         """make it as big as possible within the wall"""
-        cWall = InternalParameters.field.game.wall
+        cWall = Internal.field.game.wall
         newSceneX = cWall[3].sceneBoundingRect().right()
         newSceneY = cWall[2].sceneBoundingRect().bottom()
-        QGraphicsRectItem.setPos(self, newSceneX, newSceneY)
         tileset = self.tileset
-        avail = (cWall[2].sceneBoundingRect().width()
-            - 2 * tileset.shadowHeight()
-            - tileset.faceSize.height())
-        xScaleFactor = avail / (self.width * tileset.faceSize.width() + tileset.shadowWidth())
-        yScaleFactor = avail / (self.height * tileset.faceSize.height() + tileset.shadowHeight())
+        xAvail = cWall[1].sceneBoundingRect().left() - newSceneX
+        yAvail = cWall[0].sceneBoundingRect().top() - newSceneY
+        shadowHeight = tileset.shadowHeight()
+        shadowWidth = tileset.shadowWidth()
+        if self.showShadows:
+            # this should use the real shadow values from the wall because the wall
+            # tiles are smaller than those in the CourtBoard but this should be
+            # good enough
+            newSceneX -= shadowHeight / 2 if 'W' in self.lightSource else 0
+            newSceneY -= shadowWidth / 2 if 'N' in self.lightSource else 0
+            xAvail -= shadowHeight if 'E' in self.lightSource else 0
+            yAvail -= shadowWidth if 'S' in self.lightSource else 0
+        xNeeded = self.width * tileset.faceSize.width()
+        yNeeded = self.height * tileset.faceSize.height()
+        xScaleFactor = xAvail / xNeeded
+        yScaleFactor = yAvail / yNeeded
+        QGraphicsRectItem.setPos(self, newSceneX, newSceneY)
         Board.setScale(self, min(xScaleFactor, yScaleFactor))
         for tile in self.tiles:
             tile.board.placeTile(tile)
 
 class SelectorBoard(CourtBoard):
     """a board containing all possible tiles for selection"""
-    # pylint: disable=R0904
-    # pylint we have more than 40 public methods, shame on me!
+    # pylint: disable=too-many-public-methods
     def __init__(self):
         CourtBoard.__init__(self, 9, 5)
-        self.setAcceptDrops(True)
         self.lastReceived = None
         self.allSelectorTiles = []
 
+    def checkTiles(self):
+        """does not apply"""
+        pass
+
     def load(self, game):
         """load the tiles according to game.ruleset"""
         for tile in self.tiles:
             tile.setBoard(None)
         self.tiles = []
-        self.allSelectorTiles = list(Tile(x) for x in elements.all(game.ruleset))
+        self.allSelectorTiles = list(UITile(x) for x in elements.all(game.ruleset))
         self.refill()
 
     def refill(self):
         """move all tiles back into the selector"""
         with Animated(False):
-            for tile in self.allSelectorTiles:
-                tile.element = tile.element.lower()
-                self.__placeAvailable(tile)
-                tile.dark = False
-                tile.focusable = True
-            self.focusTile = self.tilesByElement('c1')[0]
+            for tileItem in self.allSelectorTiles:
+                tileItem.tile = Tile(tileItem.tile.lower())
+                self.__placeAvailable(tileItem)
+                tileItem.dark = False
+                tileItem.focusable = True
+            self.focusTile = self.tilesByElement(Tile('c1'))[0]
 
-    # pylint: disable=R0201
-    # pylint we know this could be static
+    # pylint: disable=no-self-use
     def name(self):
         """for debugging messages"""
         return 'selector'
 
+    def dragMoveEvent(self, event):
+        """allow dropping only from handboards, not from self"""
+        tile = event.mimeData().tile
+        event.setAccepted(tile.board != self)
+
     def dropEvent(self, event):
         """drop a tile into the selector"""
-        mime = event.mimeData()
-        self.dropHere(mime.tile, mime.meld)
+        tile = event.mimeData().tile
+        self.dropTile(tile)
         event.accept()
 
-    def receive(self, tile=None, meld=None):
-        """self receives tiles"""
-        tiles = [tile] if tile else meld.tiles
-        senderHand = tiles[0].board
-        assert senderHand != self
-        senderHand.removing(tile, meld)
-        self.lastReceived = tiles[0]
-        for myTile in tiles:
+    def dropTile(self, tile, lowerHalf=False): # pylint: disable=unused-argument
+        """drop tile or meld into selector board"""
+        senderHand = tile.board
+        if senderHand == self:
+            return
+        meld = tile.board.uiMeldWithTile(tile)
+        self.lastReceived = tile
+        for myTile in meld:
             self.__placeAvailable(myTile)
             myTile.focusable = True
-        senderHand.remove(tile, meld)
+        senderHand.deselect(meld)
         (senderHand if senderHand.tiles else self).hasFocus = True
         self._noPen()
-
-    def dropHere(self, tile, meld, dummyLowerHalf=None):
-        """drop tile or meld into selector board"""
-        tile1 = tile or meld[0]
-        if tile1.board != self:
-            self.receive(tile, meld)
         animate()
 
-    def removing(self, tile=None, meld=None):
-        """we are going to lose those tiles or melds"""
-        # pylint: disable=W0613
-        tiles = [tile] if tile else meld.tiles
-        if not self.focusTile in tiles:
-            return
-        focusCandidates = [x for x in self._focusableTiles() if x not in tiles or x == self.focusTile]
-        focusCandidates.append(focusCandidates[0])
-        self.focusTile = focusCandidates[focusCandidates.index(self.focusTile)+1]
+    def assignUITiles(self, uiTile, meld):
+        """generate a UIMeld. First uiTile is given, the rest should be as defined by meld"""
+        assert isinstance(uiTile, UITile), uiTile
+        result = UIMeld(uiTile)
+        for tile in meld[1:]:
+            baseTiles = list(x for x in self.tilesByElement(tile.lower()) if x not in result)
+            result.append(baseTiles[0])
+        return result
 
-    def remove(self, tile=None, meld=None):
-        """Default: nothing to do after something has been removed"""
-        # pylint: disable=W0613
+    def deselect(self, meld):
+        """we are going to lose those tiles or melds"""
 
     def __placeAvailable(self, tile):
         """place the tile in the selector at its place"""
@@ -638,21 +689,21 @@ class SelectorBoard(CourtBoard):
         offsets = {'d': (3, 6, 'bgr'), 'f': (4, 5, 'eswn'), 'y': (4, 0, 'eswn'),
             'w': (3, 0, 'eswn'), 'b': (1, 0, '123456789'), 's': (2, 0, '123456789'),
             'c': (0, 0, '123456789')}
-        row, baseColumn, order = offsets[tile.element[0].lower()]
-        column = baseColumn + order.index(tile.element[1])
+        row, baseColumn, order = offsets[tile.tile[0].lower()]
+        column = baseColumn + order.index(tile.tile[1])
         tile.dark = False
         tile.setBoard(self, column, row)
 
-    def meldVariants(self, tile):
-        """returns a list of possible variants based on tile."""
-        # pylint: disable=R0914
-        # pylint too many local variables
-        wantedTileName = tile.element
+    def meldVariants(self, uiTile, lowerHalf):
+        """returns a list of possible variants based on meld. Those are logical melds."""
+        # pylint: disable=too-many-locals
+        assert isinstance(uiTile, UITile)
+        wantedTile = uiTile.tile
         for selectorTile in self.tiles:
-            selectorTile.element = selectorTile.element.lower()
-        lowerName = wantedTileName.lower()
-        upperName = wantedTileName.capitalize()
-        if wantedTileName.istitle():
+            selectorTile.tile = selectorTile.tile.lower()
+        lowerName = wantedTile.lower()
+        upperName = wantedTile.upper()
+        if lowerHalf:
             scName = upperName
         else:
             scName = lowerName
@@ -663,14 +714,14 @@ class SelectorBoard(CourtBoard):
         if baseTiles >= 3:
             variants.append(scName * 3)
         if baseTiles == 4:
-            if wantedTileName.istitle():
+            if lowerHalf:
                 variants.append(lowerName + upperName * 2 + lowerName)
             else:
                 variants.append(lowerName * 4)
                 variants.append(lowerName * 3 + upperName)
-        if not tile.isHonor() and tile.element[-1] < '8':
-            chow2 = chiNext(wantedTileName, 1)
-            chow3 = chiNext(wantedTileName, 2)
+        if not wantedTile.isHonor() and wantedTile[1] < '8':
+            chow2 = chiNext(wantedTile, 1)
+            chow3 = chiNext(wantedTile, 2)
             chow2 = self.tilesByElement(chow2.lower())
             chow3 = self.tilesByElement(chow3.lower())
             if chow2 and chow3:
@@ -678,26 +729,14 @@ class SelectorBoard(CourtBoard):
                 baseValue = ord(scName[1])
                 varStr = '%s%s%s%s%s' % (scName, baseChar, chr(baseValue+1), baseChar, chr(baseValue+2))
                 variants.append(varStr)
-        result = [Meld(x) for x in variants]
-        for meld in result:
-            meld.tiles = [tile]
-            for idx, pair in enumerate(meld.pairs[1:]):
-                baseTiles = list(x for x in self.tilesByElement(pair.lower()) if x != tile)
-                meld.tiles.append(baseTiles[0 if meld.isChow() else idx])
-        return result
-
+        return [Meld(x) for x in variants]
 
 class MimeData(QMimeData):
-    """we also want to pass a reference to the moved tile"""
-    def __init__(self, tile=None, meld=None):
-        assert bool(tile) != bool(meld)
+    """we also want to pass a reference to the moved meld"""
+    def __init__(self, uiTile):
         QMimeData.__init__(self)
-        self.tile = tile
-        self.meld = meld
-        if self.tile:
-            self.setText(tile.element)
-        else:
-            self.setText(meld.joined)
+        self.tile = uiTile
+        self.setText(str(uiTile))
 
 class FittingView(QGraphicsView):
     """a graphics view that always makes sure the whole scene is visible"""
@@ -718,11 +757,11 @@ class FittingView(QGraphicsView):
         self.dragObject = None
         self.setFocus()
 
-    def wheelEvent(self, event):  # pylint: disable=R0201
+    def wheelEvent(self, event):  # pylint: disable=no-self-use
         """we do not want scrolling for the scene view.
         Instead scrolling down changes perspective like in kmahjongg"""
         if event.orientation() == Qt.Vertical and event.delta() < 0:
-            InternalParameters.field.changeAngle()
+            Internal.field.changeAngle()
         # otherwise do not call ignore() because we do want
         # to consume this
 
@@ -740,28 +779,27 @@ class FittingView(QGraphicsView):
                 grandpa.applySettings()
                 # resize background:
                 grandpa.backgroundName = grandpa.backgroundName
-        if InternalParameters.scaleScene and self.scene():
+        if Internal.scaleScene and self.scene():
             self.fitInView(self.scene().itemsBoundingRect(), Qt.KeepAspectRatio)
         self.setFocus()
 
     def __matchingTile(self, position, item):
         """is position in the clickableRect of this tile?"""
-        if not isinstance(item, GraphicsTileItem):
+        if not isinstance(item, UITile):
             return False
         itemPos = item.mapFromScene(self.mapToScene(position))
-        return item.tile.board.tileFaceRect().contains(itemPos)
+        return item.board.tileFaceRect().contains(itemPos)
 
     def tileAt(self, position):
         """find out which tile is clickable at this position. Always
         returns a list. If there are several tiles above each other,
         return all of them, highest first"""
-        allTiles = [x for x in self.items(position) if isinstance(x, GraphicsTileItem)]
+        allTiles = [x for x in self.items(position) if isinstance(x, UITile)]
         items = [x for x in allTiles if self.__matchingTile(position, x)]
         if not items:
             return None
-        items = [x.tile for x in items]
         for item in items[:]:
-            for other in [x.tile for x in allTiles]:
+            for other in allTiles:
                 if (other.xoffset, other.yoffset) == (item.xoffset, item.yoffset):
                     if other.level > item.level:
                         items.append(other)
@@ -777,13 +815,13 @@ class FittingView(QGraphicsView):
             tile = tiles[0]
             board = tile.board
             isRemote = board.isHandBoard and board.player and not board.player.game.isScoringGame()
-            if board.isHandBoard and not isRemote:
-                tile = tile.board.meldWithTile(tile)[0]
+            if board.isHandBoard and not isRemote and not tile.isBonus():
+                tile = tile.board.uiMeldWithTile(tile)[0]
             if tile.focusable:
                 board.focusTile = tile
                 board.hasFocus = True
                 if isRemote:
-                    InternalParameters.field.clientDialog.buttons[0].setFocus()
+                    Internal.field.clientDialog.buttons[0].setFocus()
                 self.tilePressed = tile
             else:
                 event.ignore()
@@ -803,26 +841,21 @@ class FittingView(QGraphicsView):
         if tilePressed:
             board = tilePressed.board
             if board and board.tileDragEnabled:
-                selBoard = InternalParameters.field.selectorBoard
-                selBoard.setAcceptDrops(tilePressed.board != selBoard)
-                tile, meld = board.dragObject(tilePressed)
-                self.dragObject = self.drag(tile, meld)
+                self.dragObject = self.drag(tilePressed)
                 self.dragObject.exec_(Qt.MoveAction)
                 self.dragObject = None
                 return
         return QGraphicsView.mouseMoveEvent(self, event)
 
-    def drag(self, tile=None, meld=None):
+    def drag(self, tile):
         """returns a drag object"""
         drag = QDrag(self)
-        mimeData = MimeData(tile, meld)
+        mimeData = MimeData(tile)
         drag.setMimeData(mimeData)
-        tile = tile or meld[0]
-        graphics = tile.graphics
-        tRect = graphics.boundingRect()
+        tRect = tile.boundingRect()
         tRect = self.viewportTransform().mapRect(tRect)
-        pmapSize = QSize(tRect.width() * graphics.scale(), tRect.height() * graphics.scale())
-        pMap = graphics.pixmapFromSvg(pmapSize)
+        pmapSize = QSize(tRect.width() * tile.scale, tRect.height() * tile.scale)
+        pMap = tile.pixmapFromSvg(pmapSize)
         drag.setPixmap(pMap)
         drag.setHotSpot(QPoint(pMap.width()/2, pMap.height()/2))
         return drag
@@ -871,14 +904,17 @@ class DiscardBoard(CourtBoard):
         CourtBoard.__init__(self, 11, 9)
         self.__places = None
         self.lastDiscarded = None
-        self.setAcceptDrops(True)
 
-    # pylint: disable=R0201
-    # pylint we do not want this to be staticmethod
-    def name(self):
+    @staticmethod
+    def name(): # pylint: disable=arguments-differ
         """to be used in debug output"""
         return "discardBoard"
 
+    def hide(self):
+        """remove all tile references so they can be garbage collected"""
+        self.lastDiscarded = None
+        CourtBoard.hide(self)
+
     def setRandomPlaces(self, randomGenerator):
         """precompute random positions"""
         self.__places = [(x, y) for x in range(self.width) for y in range(self.height)]
@@ -886,7 +922,7 @@ class DiscardBoard(CourtBoard):
 
     def discardTile(self, tile):
         """add tile to a random position"""
-        assert isinstance(tile, Tile)
+        assert isinstance(tile, UITile)
         tile.setBoard(self, *self.__places.pop(0))
         tile.dark = False
         tile.focusable = False
@@ -896,12 +932,12 @@ class DiscardBoard(CourtBoard):
 
     def dropEvent(self, event):
         """drop a tile into the discard board"""
-        # now that tiles are top level scene items, maybe drag them
+        # TODO: now that tiles are top level scene items, maybe drag them
         # directly. Draggings melds: QGraphicsItemGroup?
-        mime = event.mimeData()
-        graphics = mime.tile.graphics
-        graphics.setPos(event.scenePos() - graphics.boundingRect().center())
-        InternalParameters.field.clientDialog.selectButton(Message.Discard)
+        tile = event.mimeData().tile
+        assert isinstance(tile, UITile), tile
+        tile.setPos(event.scenePos() - tile.boundingRect().center())
+        Internal.field.clientDialog.selectButton(Message.Discard)
         event.accept()
         self._noPen()
 
@@ -927,7 +963,7 @@ class MJScene(QGraphicsScene):
            wait until the main screen has been built
            click with the mouse into the middle of that window
            press left arrow key
-           this will violate the assertion in GraphicsTileItem.keyPressEvent """
+           this will violate the assertion in UITile.keyPressEvent """
         prev = self.focusItem()
         QGraphicsScene.focusInEvent(self, event)
         if prev and bool(prev.flags() & QGraphicsItem.ItemIsFocusable) and prev != self.focusItem():
@@ -935,55 +971,54 @@ class MJScene(QGraphicsScene):
 
     def __focusRectVisible(self):
         """should we show it?"""
-        game = InternalParameters.field.game
+        game = Internal.field.game
         board = self._focusBoard
         return bool(not self.__disableFocusRect
                 and board
+                and board.isVisible()
+                and board.isEnabled()
                 and board.hasFocus
                 and board.focusTile
                 and game
                 and not game.autoPlay)
 
-    @apply
-    def disableFocusRect():
+    @property
+    def disableFocusRect(self):
         """suppress focusrect"""
-        def fget(self):
-            # pylint: disable=W0212
-            return self.__disableFocusRect
-        def fset(self, value):
-            # pylint: disable=W0212
-            # always place or hide, even if value does not change
-            self.__disableFocusRect = value
-            if value:
-                self.focusRect.hide()
-            else:
-                self.placeFocusRect()
-        return property(**locals())
+        return self.__disableFocusRect
 
-    @apply
-    def focusBoard():
+    @disableFocusRect.setter
+    def disableFocusRect(self, value):
+        """always place or hide, even if value does not change"""
+        self.__disableFocusRect = value
+        if value:
+            self.focusRect.hide()
+        else:
+            self.placeFocusRect()
+
+    @property
+    def focusBoard(self):
         """get / set the board that has its focusRect shown"""
-        def fget(self):
-            # pylint: disable=W0212
-            return self._focusBoard
-        def fset(self, board):
-            # pylint: disable=W0212
-            self._focusBoard = board
-            focusTile = board.focusTile if board else None
-            if focusTile:
-                focusTile.graphics.setFocus()
-                self.placeFocusRect()
-            self.focusRect.setVisible(self.__focusRectVisible())
-        return property(**locals())
+        return self._focusBoard
+
+    @focusBoard.setter
+    def focusBoard(self, board):
+        """get / set the board that has its focusRect shown"""
+        self._focusBoard = board
+        focusTile = board.focusTile if board else None
+        if focusTile:
+            focusTile.setFocus()
+            self.placeFocusRect()
+        self.focusRect.setVisible(self.__focusRectVisible())
 
     def placeFocusRect(self):
         """show a blue rect around tile"""
         board = self._focusBoard
-        if isAlive(board) and self.__focusRectVisible():
+        if board and board.isVisible() and board.focusTile:
             rect = board.tileFaceRect()
             rect.setWidth(rect.width()*board.focusRectWidth())
             self.focusRect.setRect(rect)
-            self.focusRect.setPos(board.focusTile.graphics.pos())
+            self.focusRect.setPos(board.focusTile.pos)
             self.focusRect.setRotation(board.sceneRotation())
             self.focusRect.setScale(board.scale())
             self.focusRect.show()
@@ -991,12 +1026,12 @@ class MJScene(QGraphicsScene):
             self.focusRect.hide()
 
     def graphicsTileItems(self):
-        """returns all GraphicsTileItems in the scene"""
-        return (x for x in self.items() if isinstance(x, GraphicsTileItem))
+        """returns all UITile in the scene"""
+        return (x for x in self.items() if isinstance(x, UITile))
 
     def nonTiles(self):
         """returns all other items in the scene"""
-        return (x for x in self.items() if not isinstance(x, GraphicsTileItem))
+        return (x for x in self.items() if not isinstance(x, UITile))
 
     def removeTiles(self):
         """remove all tiles from scene"""
