@@ -22,9 +22,10 @@ Read the user manual for a description of the interface to this scoring engine
 """
 
 from util import logDebug
-from meld import Meld, meldKey, meldsContent, Pairs, CONCEALED
+from tile import Tile, elements
+from meld import Meld, meldKey, meldsContent, CONCEALED
 from rule import Score, Ruleset
-from common import elements, Debug
+from common import Debug
 
 class UsedRule(object):
     """use this in scoring, never change class Rule.
@@ -33,11 +34,19 @@ class UsedRule(object):
         self.rule = rule
         self.meld = meld
 
+    def __str__(self):
+        result = self.rule.name
+        if self.meld:
+            result += ' ' + str(self.meld)
+        return result
+
+    def __repr__(self):
+        return 'UsedRule(%s)' % str(self)
+
 class Hand(object):
     """represent the hand to be evaluated"""
 
-    # pylint: disable=R0902
-    # pylint we need more than 10 instance attributes
+    # pylint: disable=too-many-instance-attributes
 
     cache = dict()
     misses = 0
@@ -55,6 +64,8 @@ class Hand(object):
     @staticmethod
     def cached(ruleset, string, computedRules=None, robbedTile=None):
         """since a Hand instance is never changed, we can use a cache"""
+        if computedRules is not None and not isinstance(computedRules, list):
+            computedRules = list([computedRules])
         cRuleHash = '&&'.join([rule.name for rule in computedRules]) if computedRules else 'None'
         if isinstance(ruleset, Hand):
             cacheId = id(ruleset.player or ruleset.ruleset)
@@ -78,7 +89,7 @@ class Hand(object):
         """evaluate string using ruleset. rules are to be applied in any case.
         ruleset can be Hand, Game or Ruleset."""
         # silence pylint. This method is time critical, so do not split it into smaller methods
-        # pylint: disable=R0902,R0914,R0912,R0915
+        # pylint: disable=too-many-instance-attributes,too-many-branches,too-many-statements
         if isinstance(ruleset, Hand):
             self.ruleset = ruleset.ruleset
             self.player = ruleset.player
@@ -91,6 +102,8 @@ class Hand(object):
             self.ruleset = self.player.game.ruleset
         self.string = string
         self.robbedTile = robbedTile
+        if computedRules is not None and not isinstance(computedRules, list):
+            computedRules = list([computedRules])
         self.computedRules = computedRules or []
         self.__won = False
         self.mjStr = ''
@@ -121,12 +134,13 @@ class Hand(object):
         self.mjStr = ' '.join(mjStrings)
         self.__lastTile = self.__lastSource = self.__announcements = ''
         self.__lastMeld = 0
+        self.__lastMelds = []
         self.hiddenMelds = []
         self.declaredMelds = []
         self.melds = []
         tileString = ' '.join(tileStrings)
         self.bonusMelds, tileString = self.__separateBonusMelds(tileString)
-        self.tileNames = Pairs(tileString.replace(' ','').replace('R', ''))
+        self.tileNames = Meld(tileString.replace(' ','').replace('R', ''))
         self.tileNames.sort()
         self.values = ''.join(x[1] for x in self.tileNames)
         self.suits = set(x[0].lower() for x in self.tileNames)
@@ -134,7 +148,9 @@ class Hand(object):
         self.dragonMelds, self.windMelds = self.__computeDragonWindMelds(tileString)
         self.__separateMelds(tileString)
         self.hiddenMelds = sorted(self.hiddenMelds, key=meldKey)
-        self.inHand = sum((x.pairs for x in self.hiddenMelds), [])
+        self.tilesInHand = sum(self.hiddenMelds, [])
+        for tile in self.tilesInHand:
+            assert isinstance(tile, Tile), self.tilesInHand
         self.sortedMeldsContent = meldsContent(self.melds)
         if self.bonusMelds:
             self.sortedMeldsContent += ' ' + meldsContent(self.bonusMelds)
@@ -143,6 +159,8 @@ class Hand(object):
         self.score = None
         oldWon = self.won
         self.__applyRules()
+        if len(self.lastMelds) > 1:
+            self.__applyBestLastMeld()
         if self.won != oldWon:
             # if not won after all, this might be a long hand.
             # So we might even have to unapply meld rules and
@@ -152,64 +170,61 @@ class Hand(object):
             # let you declare an invalid mah jongg
             self.__applyRules()
 
-    @apply
-    def lastTile():
-        """compute and cache, readonly"""
-        def fget(self):
-            # pylint: disable=W0212
-            if self.__lastTile == '':
-                self.__setLastTile()
-            return self.__lastTile
-        return property(**locals())
+    def hasTiles(self):
+        """tiles are assigned to this hand"""
+        return self.tileNames or self.bonusMelds
 
-    @apply
-    def lastSource():
+    @property
+    def lastTile(self):
         """compute and cache, readonly"""
-        def fget(self):
-            # pylint: disable=W0212
-            if self.__lastTile == '':
-                self.__setLastTile()
-            return self.__lastSource
-        return property(**locals())
+        if self.__lastTile == '':
+            self.__setLastTile()
+        return self.__lastTile
 
-    @apply
-    def announcements():
+    @property
+    def lastSource(self):
         """compute and cache, readonly"""
-        def fget(self):
-            # pylint: disable=W0212
-            if self.__lastTile == '':
-                self.__setLastTile()
-            return self.__announcements
-        return property(**locals())
+        if self.__lastTile == '':
+            self.__setLastTile()
+        return self.__lastSource
 
-    @apply
-    def lastMeld():
+    @property
+    def announcements(self):
         """compute and cache, readonly"""
-        def fget(self):
-            # pylint: disable=W0212
-            if self.__lastMeld == 0:
-                self.__setLastMeld()
-            return self.__lastMeld
-        return property(**locals())
+        if self.__lastTile == '':
+            self.__setLastTile()
+        return self.__announcements
 
-    @apply
-    def won(): # pylint: disable=E0202
+    @property
+    def lastMeld(self):
+        """compute and cache, readonly"""
+        if self.__lastMeld == 0:
+            self.__setLastMeld()
+        return self.__lastMeld
+
+    @property
+    def lastMelds(self):
+        """compute and cache, readonly"""
+        if self.__lastMeld == 0:
+            self.__setLastMeld()
+        return self.__lastMelds
+
+    @property
+    def won(self):
         """have we been modified since load or last save?
         The "won" value is set to True when instantiating the hand,
         according to the mMx in the init string. Later on, it may
         only be cleared."""
-        def fget(self):
-            # pylint: disable=W0212
-            return self.__won
-        def fset(self, value):
-            # pylint: disable=W0212
-            # pylint: disable=W0142
-            value = bool(value)
-            assert not value
-            self.__won = value
-            self.string = self.string.replace(' M', ' m')
-            self.mjStr = self.mjStr.replace(' M', ' m')
-        return property(**locals())
+        return self.__won
+
+    @won.setter
+    def won(self, value):
+        """must never change to True"""
+        value = bool(value)
+        assert not value
+        self.__won = value
+        self.string = self.string.replace(' M', ' m')
+        self.mjStr = self.mjStr.replace(' M', ' m')
 
     def debug(self, msg, btIndent=None):
         """try to use Game.debug so we get a nice prefix"""
@@ -229,6 +244,11 @@ class Hand(object):
         if self.__hasExclusiveRules():
             return
         self.score = self.__totalScore()
+
+        # do the rest only if we know all tiles of the hand
+        if 'Xy' in self.string:
+            self.won = False    # we do not know better
+            return
         if self.won:
             matchingMJRules = self.__maybeMahjongg()
             if not matchingMJRules:
@@ -276,15 +296,15 @@ class Hand(object):
                 part = part[1:]
                 if len(part) > 2:
                     self.__lastMeld = Meld(part[2:])
-                self.__lastTile = part[:2]
+                self.__lastTile = Tile(part[:2])
             elif part[0] == 'M':
                 if len(part) > 3:
                     self.__lastSource = part[3]
                     if len(part) > 4:
                         self.__announcements = part[4:]
         if self.__lastTile:
-            assert self.__lastTile in self.tileNames, 'lastTile %s is not in tiles %s' % (
-                self.__lastTile, ' '.join(self.tileNames))
+            assert self.__lastTile in self.tileNames, 'lastTile %s is not in tiles %s, mjStr=%s' % (
+                self.__lastTile, ' '.join(self.tileNames), self.mjStr)
             if self.__lastSource == 'k':
                 assert self.tileNames.count(self.__lastTile.lower()) + \
                     self.tileNames.count(self.__lastTile.capitalize()) == 1, \
@@ -293,16 +313,44 @@ class Hand(object):
 
 
     def __setLastMeld(self):
-        """sets best last meld"""
+        """sets the shortest possible last meld. This is
+        not yet the final choice, see __applyBestLastMeld"""
         self.__lastMeld = None
         if self.lastTile and self.won:
             if hasattr(self.mjRule.function, 'computeLastMelds'):
-                lastMelds = self.mjRule.function.computeLastMelds(self)
-                if lastMelds:
+                self.__lastMelds = self.mjRule.function.computeLastMelds(self)
+                if self.__lastMelds:
                     # syncHandBoard may return nothing
-                    self.__lastMeld = lastMelds[0] # TODO: use the one with highest score
+                    if len(self.__lastMelds) == 1:
+                        self.__lastMeld = self.__lastMelds[0]
+                    else:
+                        totals = sorted((len(x), idx) for idx, x in enumerate(self.__lastMelds))
+                        self.__lastMeld = self.__lastMelds[totals[0][1]]
             if not self.__lastMeld:
                 self.__lastMeld = Meld([self.lastTile])
+                self.__lastMelds = [self.__lastMeld]
+
+    def __applyBestLastMeld(self):
+        """select the last meld giving the highest score (only winning variants)"""
+        assert len(self.lastMelds) > 1
+        totals = []
+        prev = self.lastMeld
+        for lastMeld in self.lastMelds:
+            self.__lastMeld = lastMeld
+            self.__applyRules()
+            totals.append((self.won, self.__totalScore().total(), lastMeld))
+        if any(x[0] for x in totals): # if any won
+            totals = list(x[1:] for x in totals if x[0]) # remove lost variants
+            totals = sorted(totals) # sort by totalScore
+            maxScore = totals[-1][0]
+            totals = list(x[1] for x in totals if x[0] == maxScore)
+            # now we have a list of only lastMelds reaching maximum score
+            if prev not in totals or self.__lastMeld not in totals:
+                if Debug.explain and prev not in totals:
+                    if not self.player or not self.player.game.belongsToRobotPlayer():
+                        self.debug('replaced last meld %s with %s' % (prev, totals[0]))
+                self.__lastMeld = totals[0]
+                self.__applyRules()
 
     def __sub__(self, tiles):
         """returns a copy of self minus tiles. Case of tiles (hidden
@@ -310,38 +358,37 @@ class Hand(object):
         but found in an exposed meld, this meld will be hidden with
         the tile removed from it. Exposed melds of length<3 will also
         be hidden."""
-        # pylint: disable=R0912
-        # pylint says too many branches
+        # pylint: disable=too-many-branches
         if not isinstance(tiles, list):
             tiles = list([tiles])
-        hidden = 'R' + ''.join(self.inHand)
+        hidden = 'R' + ''.join(self.tilesInHand)
         # exposed is a deep copy of declaredMelds. If lastMeld is given, it
         # must be first in the list.
         exposed = (Meld(x) for x in self.declaredMelds)
         if self.lastMeld:
-            exposed = sorted(exposed, key=lambda x: (x.pairs != self.lastMeld.pairs, meldKey(x)))
+            exposed = sorted(exposed, key=lambda x: (x != self.lastMeld, meldKey(x)))
         else:
             exposed = sorted(exposed, key=meldKey)
-        bonus = sorted(Meld(x) for x in self.bonusMelds)
+        boni = sorted(self.bonusMelds)
         for tile in tiles:
-            assert isinstance(tile, str) and len(tile) == 2, 'Hand.__sub__:%s' % tiles
-            if tile.capitalize() in hidden:
-                hidden = hidden.replace(tile.capitalize(), '', 1)
-            elif tile[0] in 'fy': # bonus tile
-                for idx, meld in enumerate(bonus):
-                    if tile == meld.pairs[0]:
-                        del bonus[idx]
+            assert isinstance(tile, Tile), tiles
+            if tile.upper() in hidden:
+                hidden = hidden.replace(tile.upper(), '', 1)
+            elif tile.isBonus():
+                for idx, meld in enumerate(boni):
+                    if tile == meld[0]:
+                        del boni[idx]
                         break
             else:
                 for idx, meld in enumerate(exposed):
-                    if tile.lower() in meld.pairs:
-                        del meld.pairs[meld.pairs.index(tile.lower())]
+                    if tile.lower() in meld:
+                        del meld[meld.index(tile.lower())]
                         del exposed[idx]
                         meld.conceal()
                         hidden += meld.joined
                         break
         for idx, meld in enumerate(exposed):
-            if len(meld.pairs) < 3:
+            if len(meld) < 3:
                 del exposed[idx]
                 meld.conceal()
                 hidden += meld.joined
@@ -358,7 +405,7 @@ class Hand(object):
                     continue
                 newParts.append(part)
             mjStr = ' '.join(newParts)
-        newString = ' '.join([hidden, meldsContent(exposed), meldsContent(bonus), mjStr])
+        newString = ' '.join([hidden, meldsContent(exposed), meldsContent(boni), mjStr])
         return Hand.cached(self, newString, self.computedRules)
 
     def manualRuleMayApply(self, rule):
@@ -426,42 +473,40 @@ class Hand(object):
                 splits = rule.apply(rest)
             if len(splits) == 0:
                 break
-        if len(splits) == 1 :
-            assert Meld(splits[0]).isValid()   # or the splitRules are wrong
         return melds
 
+    def __recurse(self, cVariants, foundMelds, rest, maxPairs, group):
+        """build the variants recursively"""
+        melds = []
+        for value in set(rest):
+            intValue = int(value)
+            if rest.count(value) == 3:
+                melds.append([value] * 3)
+            elif rest.count(value) == 2:
+                melds.append([value] * 2)
+            if rest.count(str(intValue + 1)) and rest.count(str(intValue + 2)):
+                melds.append([value, str(intValue+1), str(intValue+2)])
+        pairsFound = sum(len(x) == 2 for x in foundMelds)
+        for meld in (m for m in melds if len(m) !=2 or pairsFound < maxPairs):
+            restCopy = rest[:]
+            for value in meld:
+                restCopy.remove(value)
+            newMelds = foundMelds[:]
+            newMelds.append(meld)
+            if restCopy:
+                self.__recurse(cVariants, newMelds, restCopy, maxPairs, group)
+            else:
+                for idx, newMeld in enumerate(newMelds):
+                    newMelds[idx] = ''.join(group+x for x in newMeld)
+                cVariants.append(' '.join(sorted(newMelds )))
+
     def genVariants(self, original0, maxPairs=1):
         """generates all possible meld variants out of original
         where original is a list of tile values like ['1','1','2']"""
-        color = original0[0][0]
+        group = original0[0][0]
         original = [x[1] for x in original0]
-        def recurse(cVariants, foundMelds, rest):
-            """build the variants recursively"""
-            values = set(rest)
-            melds = []
-            for value in values:
-                intValue = int(value)
-                if rest.count(value) == 3:
-                    melds.append([value] * 3)
-                elif rest.count(value) == 2:
-                    melds.append([value] * 2)
-                if rest.count(str(intValue + 1)) and rest.count(str(intValue + 2)):
-                    melds.append([value, str(intValue+1), str(intValue+2)])
-            pairsFound = sum(len(x) == 2 for x in foundMelds)
-            for meld in (m for m in melds if len(m) !=2 or pairsFound < maxPairs):
-                restCopy = rest[:]
-                for value in meld:
-                    restCopy.remove(value)
-                newMelds = foundMelds[:]
-                newMelds.append(meld)
-                if restCopy:
-                    recurse(cVariants, newMelds, restCopy)
-                else:
-                    for idx, newMeld in enumerate(newMelds):
-                        newMelds[idx] = ''.join(color+x for x in newMeld)
-                    cVariants.append(' '.join(sorted(newMelds )))
         cVariants = []
-        recurse(cVariants, [], original)
+        self.__recurse(cVariants, [], original, maxPairs, group)
         gVariants = []
         for cVariant in set(cVariants):
             melds = [Meld(x) for x in cVariant.split()]
@@ -588,7 +633,7 @@ class Hand(object):
         windMelds = []
         for split in tileString.split():
             if split[0] == 'R':
-                pairs = Pairs(split[1:])
+                pairs = Meld(split[1:])
                 for lst, tiles in ((windMelds, elements.wINDS), (dragonMelds, elements.dRAGONS)):
                     for tile in tiles:
                         count = pairs.count(tile)
@@ -605,8 +650,8 @@ class Hand(object):
         """keep them separate. One meld per bonus tile. Others depend on that."""
         result = []
         if 'f' in tileString or 'y' in tileString:
-            for pair in Pairs(tileString.replace(' ','').replace('R', '')):
-                if pair[0] in 'fy':
+            for pair in Meld(tileString.replace(' ','').replace('R', '')):
+                if pair.isBonus():
                     result.append(Meld(pair))
                     tileString = tileString.replace(pair, '', 1)
         return result, tileString
@@ -630,9 +675,6 @@ class Hand(object):
             rest = sorted([rest[x:x+2] for x in range(0, len(rest), 2)])
             self.__split(rest)
         self.melds = sorted(self.melds, key=meldKey)
-        for meld in self.melds:
-            if not meld.isValid():
-                raise Exception('%s has an invalid meld: %s' % (self.string, meld.joined))
         self.__categorizeMelds()
 
     def picking(self, tileName):
@@ -673,13 +715,18 @@ class Hand(object):
 
     def explain(self):
         """explain what rules were used for this hand"""
-        result = [x.rule.explain() for x in self.usedRules
+        if self.player:
+            usedRules = self.player.sortRulesByX(self.usedRules)
+        else:
+            # scoringtest
+            usedRules = self.usedRules
+        result = [x.rule.explain(x.meld) for x in usedRules
             if x.rule.score.points]
-        result.extend([x.rule.explain() for x in self.usedRules
+        result.extend([x.rule.explain(x.meld) for x in usedRules
             if x.rule.score.doubles])
-        result.extend([x.rule.explain() for x in self.usedRules
+        result.extend([x.rule.explain(x.meld) for x in usedRules
             if not x.rule.score.points and not x.rule.score.doubles])
-        if any(x.rule.debug for x in self.usedRules):
+        if any(x.rule.debug for x in usedRules):
             result.append(str(self))
         return result
 
@@ -699,3 +746,7 @@ class Hand(object):
     def __str__(self):
         """hand as a string"""
         return u' '.join([self.sortedMeldsContent, self.mjStr])
+
+    def __repr__(self):
+        """the default representation"""
+        return 'Hand(%s)' % str(self)
