@@ -18,7 +18,7 @@ along with this program if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 """
 
-import datetime, random
+import datetime
 
 from kde import KIcon, KDialogButtonBox
 from PyQt4.QtCore import Qt, QVariant, \
@@ -31,13 +31,12 @@ from kde import KApplication
 
 from genericdelegates import RichTextColumnDelegate
 
-from util import logWarning, m18n, m18nc, m18nE, logDebug
+from util import m18n, m18nc, m18nE, logDebug
 from statesaver import StateSaver
 from rule import Ruleset
 from guiutil import ListComboBox, MJTableView
 from differ import RulesetDiffer
-from common import InternalParameters, Debug
-from client import ClientTable
+from common import Internal, Debug
 from modeltest import ModelTest
 from chat import ChatMessage, ChatWindow
 
@@ -46,8 +45,9 @@ class TablesModel(QAbstractTableModel):
     def __init__(self, tables, parent = None):
         super(TablesModel, self).__init__(parent)
         self.tables = tables
+        assert isinstance(tables, list)
 
-    def headerData(self, section, orientation, role=Qt.DisplayRole): # pylint: disable=R0201
+    def headerData(self, section, orientation, role=Qt.DisplayRole): # pylint: disable=no-self-use
         """show header"""
         if role == Qt.TextAlignmentRole:
             if orientation == Qt.Horizontal:
@@ -71,7 +71,7 @@ class TablesModel(QAbstractTableModel):
             return 0
         return len(self.tables)
 
-    def columnCount(self, dummyParent=None): # pylint: disable=R0201
+    def columnCount(self, dummyParent=None): # pylint: disable=no-self-use
         """for now we only have id (invisible), id (visible), players, status, ruleset.name.
         id(invisible) always holds the real id, also 1000 for suspended tables.
         id(visible) is what should be displayed."""
@@ -79,8 +79,7 @@ class TablesModel(QAbstractTableModel):
 
     def data(self, index, role=Qt.DisplayRole):
         """score table"""
-        # pylint: disable=R0912,R0914
-        # pylint too many branches
+        # pylint: disable=too-many-branches,too-many-locals
         result = QVariant()
         if role == Qt.TextAlignmentRole:
             if index.column() == 0:
@@ -89,11 +88,8 @@ class TablesModel(QAbstractTableModel):
                 result = QVariant(int(Qt.AlignLeft|Qt.AlignVCenter))
         if index.isValid() and (0 <= index.row() < len(self.tables)):
             table = self.tables[index.row()]
-            if role == Qt.DisplayRole and index.column() == 1:
+            if role == Qt.DisplayRole and index.column() in (0, 1):
                 result = QVariant(table.tableid)
-            elif role == Qt.DisplayRole and index.column() == 0:
-                if not table.status.startswith('Suspended'):
-                    result = QVariant(table.tableid)
             elif role == Qt.DisplayRole and index.column() == 2:
                 players = []
                 zipped = zip(table.playerNames, table.playersOnline)
@@ -112,9 +108,9 @@ class TablesModel(QAbstractTableModel):
                 names = ''.join(players)
                 result = QVariant(names)
             elif role == Qt.DisplayRole and index.column() == 3:
-                status = table.status
-                if status.startswith('Suspended'):
-                    dateVal = ' ' + datetime.datetime.strptime(status.replace('Suspended', ''),
+                status = table.status()
+                if table.suspendedAt:
+                    dateVal = ' ' + datetime.datetime.strptime(table.suspendedAt,
                         '%Y-%m-%dT%H:%M:%S').strftime('%c').decode('utf-8')
                     status = 'Suspended'
                 else:
@@ -122,7 +118,7 @@ class TablesModel(QAbstractTableModel):
                 result = QVariant(m18nc('table status', status) + dateVal)
             elif index.column() == 4:
                 if role == Qt.DisplayRole:
-                    result = QVariant(m18n(table.ruleset.name))
+                    result = QVariant(m18n((table.myRuleset if table.myRuleset else table.ruleset).name))
                 elif role == Qt.ForegroundRole:
                     palette = KApplication.palette()
                     color = palette.windowText() if table.myRuleset else 'red'
@@ -131,7 +127,7 @@ class TablesModel(QAbstractTableModel):
 
 class SelectRuleset(QDialog):
     """a dialog for selecting a ruleset"""
-    def __init__(self, server):
+    def __init__(self, server=None):
         QDialog.__init__(self, None)
         self.setWindowTitle(m18n('Select a ruleset') + ' - Kajongg')
         self.buttonBox = KDialogButtonBox(self)
@@ -149,8 +145,7 @@ class SelectRuleset(QDialog):
 
 class TableList(QWidget):
     """a widget for viewing, joining, leaving tables"""
-    # pylint: disable=R0902
-    # pylint we have more than 10 attributes
+    # pylint: disable=too-many-instance-attributes
     def __init__(self, client):
         super(TableList, self).__init__(None)
         self.autoStarted = False
@@ -160,16 +155,16 @@ class TableList(QWidget):
         self.view = MJTableView(self)
         self.differ = None
         self.debugModelTest = None
-        self.__requestedNewTable = False
+        self.requestedNewTable = False
         self.view.setItemDelegateForColumn(2, RichTextColumnDelegate(self.view))
 
         buttonBox = QDialogButtonBox(self)
         self.newButton = buttonBox.addButton(m18nc('allocate a new table', "&New"), QDialogButtonBox.ActionRole)
         self.newButton.setIcon(KIcon("document-new"))
         self.newButton.setToolTip(m18n("Allocate a new table"))
-        self.newButton.clicked.connect(self.newTable)
+        self.newButton.clicked.connect(self.client.newTable)
         self.joinButton = buttonBox.addButton(m18n("&Join"), QDialogButtonBox.AcceptRole)
-        self.joinButton.clicked.connect(self.joinTable)
+        self.joinButton.clicked.connect(client.joinTable)
         self.joinButton.setIcon(KIcon("list-add-user"))
         self.joinButton.setToolTip(m18n("Join a table"))
         self.leaveButton = buttonBox.addButton(m18n("&Leave"), QDialogButtonBox.AcceptRole)
@@ -196,14 +191,20 @@ class TableList(QWidget):
         layout.addLayout(cmdLayout)
         self.setLayout(layout)
 
-        self.view.doubleClicked.connect(self.joinTable)
+        self.view.doubleClicked.connect(client.joinTable)
         StateSaver(self, self.view.horizontalHeader())
         self.updateButtonsForTable(None)
 
-    def hideEvent(self, dummyEvent): # pylint: disable=R0201
+    def hideEvent(self, dummyEvent): # pylint: disable=no-self-use
         """table window hides"""
-        field = InternalParameters.field
+        field = Internal.field
         field.startingGame = False
+        model = self.view.model()
+        if model:
+            for table in model.tables:
+                if table.chatWindow:
+                    table.chatWindow.hide()
+                    table.chatWindow = None
         if not field.game or field.game.client != self.client:
             # do we still need this connection?
             self.client.logout()
@@ -220,8 +221,8 @@ class TableList(QWidget):
         table = self.selectedTable()
         if not table.chatWindow:
             line = m18nE('opens a chat window')
-            msg = ChatMessage(table.tableid, table.client.username, line, isStatusMessage=True)
-            table.client.sendChat(msg).addCallback(initChat).addErrback(self.tableError)
+            msg = ChatMessage(table.tableid, table.client.name, line, isStatusMessage=True)
+            table.client.sendChat(msg).addCallback(initChat).addErrback(self.client.tableError)
         elif table.chatWindow.isVisible():
             table.chatWindow.hide()
         else:
@@ -229,12 +230,11 @@ class TableList(QWidget):
 
     def show(self):
         """prepare the view and show it"""
-        assert not InternalParameters.autoPlay
         if self.client.hasLocalServer():
             title = m18n('Local Games with Ruleset %1', self.client.ruleset.name)
         else:
-            title = m18n('Tables at %1', self.client.url)
-        self.setWindowTitle(' - '.join([self.client.username, title, 'Kajongg']))
+            title = m18n('Tables at %1', self.client.connection.url)
+        self.setWindowTitle(' - '.join([self.client.name, title, 'Kajongg']))
         self.view.hideColumn(1)
         tableCount = self.view.model().rowCount(None) if self.view.model() else 0
         self.view.showColumn(0)
@@ -247,10 +247,6 @@ class TableList(QWidget):
                 self.view.hideColumn(2)
                 self.view.hideColumn(4)
 
-    def newLocalTable(self, newId):
-        """we just got newId from the server"""
-        self.client.callServer('startGame', newId).addErrback(self.tableError)
-
     def selectTable(self, idx):
         """select table by idx"""
         self.view.selectRow(idx)
@@ -259,13 +255,13 @@ class TableList(QWidget):
     def updateButtonsForTable(self, table):
         """update button status for the currently selected table"""
         hasTable = bool(table)
-        suspended = hasTable and table.status.startswith('Suspended')
-        running = hasTable and table.status.startswith('Running')
+        suspended = hasTable and bool(table.suspendedAt)
+        running = hasTable and table.running
         suspendedLocalGame = suspended and table.gameid and self.client.hasLocalServer()
         self.joinButton.setEnabled(hasTable and
             not running and
-            not table.isOnline(self.client.username) and
-            (self.client.username in table.playerNames) == suspended)
+            not table.isOnline(self.client.name) and
+            (self.client.name in table.playerNames) == suspended)
         self.leaveButton.setVisible(not (suspendedLocalGame))
         self.compareButton.setVisible(not (suspendedLocalGame))
         self.startButton.setVisible(not suspended)
@@ -279,10 +275,11 @@ class TableList(QWidget):
             self.joinButton.setToolTip(m18n("Join a table"))
         self.leaveButton.setEnabled(hasTable and not running and not self.joinButton.isEnabled())
         self.startButton.setEnabled(not running and not suspendedLocalGame and hasTable \
-            and self.client.username == table.playerNames[0])
+            and self.client.name == table.playerNames[0])
         self.compareButton.setEnabled(hasTable and table.myRuleset is None)
         self.chatButton.setVisible(not self.client.hasLocalServer())
-        self.chatButton.setEnabled(not running and self.leaveButton.isEnabled())
+        self.chatButton.setEnabled(not running and hasTable and self.client.name in table.playerNames
+            and sum(x.startswith('Robot ') for x in table.playerNames) < 3)
         if self.chatButton.isEnabled():
             self.chatButton.setToolTip(m18n("Chat with others on this table"))
         else:
@@ -293,50 +290,6 @@ class TableList(QWidget):
         if selected.indexes():
             self.selectTable(selected.indexes()[0].row())
 
-    @staticmethod
-    def __wantedGame():
-        """find out which game we want to start on the table"""
-        result = InternalParameters.game
-        if not result or result == '0':
-            result = str(int(random.random() * 10**9))
-        InternalParameters.game = None
-        return result
-
-    def newTable(self):
-        """I am a slot"""
-        if self.client.hasLocalServer():
-            ruleset = self.client.ruleset
-        else:
-            selectDialog = SelectRuleset(self.client.host)
-            if not selectDialog.exec_():
-                return
-            ruleset = selectDialog.cbRuleset.current
-        deferred = self.client.callServer('newTable', ruleset.toList(),
-            InternalParameters.playOpen, InternalParameters.autoPlay, self.__wantedGame()).addErrback(self.tableError)
-        if self.client.hasLocalServer():
-            deferred.addCallback(self.newLocalTable)
-        self.__requestedNewTable = True
-
-    def gotTables(self, tables):
-        """got tables for first time. If we play a local game and we have no
-        suspended game, automatically start a new one"""
-        clientTables = list(ClientTable.fromList(self.client, x) for x in tables)
-        if not InternalParameters.autoPlay:
-            if self.client.hasLocalServer():
-                # when playing a local game, only show pending tables with
-                # previously selected ruleset
-                clientTables = list(x for x in clientTables if x.ruleset == self.client.ruleset)
-        if InternalParameters.autoPlay or (not clientTables and self.client.hasLocalServer()):
-            deferred = self.client.callServer('newTable', self.client.ruleset.toList(), InternalParameters.playOpen,
-                InternalParameters.autoPlay,
-                self.__wantedGame()).addErrback(self.tableError)
-            if deferred:
-                deferred.addCallback(self.newLocalTable)
-        else:
-            self.client.tables = clientTables
-            self.loadTables(clientTables)
-            self.show()
-
     def selectedTable(self):
         """returns the selected table"""
         if self.view.selectionModel():
@@ -344,12 +297,6 @@ class TableList(QWidget):
             if index.isValid() and self.view.model():
                 return self.view.model().tables[index.row()]
 
-    def joinTable(self):
-        """join a table"""
-        table = self.selectedTable()
-        self.__requestedNewTable = table.status.startswith('Suspended') # because tableid will change
-        self.client.callServer('joinTable', table.tableid).addErrback(self.tableError)
-
     def compareRuleset(self):
         """compare the ruleset of this table against ours"""
         table = self.selectedTable()
@@ -360,25 +307,12 @@ class TableList(QWidget):
         """start playing at the selected table"""
         table = self.selectedTable()
         self.startButton.setEnabled(False)
-        self.client.callServer('startGame', table.tableid).addErrback(self.tableError)
-
-    def tableError(self, err):
-        """log the twisted error"""
-        if not self.client.perspective:
-            # lost connection to server
-            for table in self.view.model().tables:
-                if table.chatWindow:
-                    table.chatWindow.hide()
-            self.hide()
-            self.client.tableList = None
-        else:
-            logWarning(err.getErrorMessage())
+        self.client.callServer('startGame', table.tableid).addErrback(self.client.tableError)
 
     def leaveTable(self):
         """leave a table"""
         table = self.selectedTable()
-        self.__requestedNewTable = table.status.startswith('Suspended') # because tableid will change
-        self.client.callServer('leaveTable', table.tableid).addErrback(self.tableError)
+        self.client.callServer('leaveTable', table.tableid).addErrback(self.client.tableError)
 
     def __keepChatWindows(self, tables):
         """copy chatWindows from the old table list which will be thrown away"""
@@ -402,11 +336,14 @@ class TableList(QWidget):
           select that again
         else:
           select first table"""
-        if self.__requestedNewTable:
-            self.__requestedNewTable = False
-            newIds = sorted(list(set(x.tableid for x in tables) - set(x.tableid for x in self.view.model().tables)))
-            if newIds:
-                return newIds[0]
+        if self.requestedNewTable:
+            self.requestedNewTable = False
+            model = self.view.model()
+            if model:
+                oldIds = set(x.tableid for x in model.tables)
+                newIds = sorted(list(set(x.tableid for x in tables) - oldIds))
+                if newIds:
+                    return newIds[0]
         if self.selectedTable():
             return self.selectedTable().tableid
         return 0
@@ -417,14 +354,12 @@ class TableList(QWidget):
         tables that also exist locally. In theory all suspended games should
         exist locally but there might have been bugs or somebody might
         have removed the local database like when reinstalling linux"""
-        if not InternalParameters.field:
+        if not Internal.field:
             return
-        if Debug.traffic:
+        if Debug.table:
             for table in tables:
-                if not table.gameid:
-                    logDebug('%s has no gameid' % table)
-                elif not table.gameExistsLocally():
-                    logDebug('%s does not exist locally' % table)
+                if table.gameid and not table.gameExistsLocally():
+                    logDebug('Table %s does not exist locally' % table)
         tables = [x for x in tables if not x.gameid or x.gameExistsLocally()]
         tables.sort(key=lambda x: x.tableid)
         preselectTableId = self.__preselectTableId(tables)
@@ -446,6 +381,6 @@ class TableList(QWidget):
             self.newButton.setFocus()
         else:
             _ = [x for x in tables if x.tableid >= preselectTableId]
-            self.selectTable(tables.index(_[0]))
+            self.selectTable(tables.index(_[0]) if _ else 0)
         self.updateButtonsForTable(self.selectedTable())
         self.view.setFocus()
